//
//  DataManager.swift
//  DocumentRepositoryMobile
//
//  Created by Tobias on 23.10.18.
//  Copyright © 2018 Tobias. All rights reserved.
//

import Foundation
import Logging

public protocol HttpResponseFilter
{
    @available(iOS 15.0.0, *)
    func handleResponse(request: Request, data: Data, dataMapper: DataMapper?, sender: Any?) async throws -> Any
}

public protocol HttpSuccessFilter
{
    @available(iOS 15.0.0, *)

    func handleResponse(data: Data, response: HTTPURLResponse, dataMapper: DataMapper?, sender: Any?) async throws -> Any
}

public protocol DataManagerDelegate
{
    func manipulateRequest(request: Request) -> Request
}

open class DataManager: NSObject, URLSessionDelegate
{
    private static let logger = Logger(label: "de.tobias.AppleLibs.requests")
    
    public enum DataManagerError: Error
    {
        case authError
    }

    public struct Host: Equatable
    {
        public let `protocol`: String
        public let host: String
        public let port: Int
        public let context: String
        
        public init(`protocol`: String, host: String, port: Int, context: String) {
            self.protocol = `protocol`
            self.host = host
            self.port = port
            self.context = context
        }
    }
    
    public typealias HostCallback = () -> Host

    private let timeout: TimeInterval
    public private(set) var hostCallback: HostCallback

    private lazy var urlSession: URLSession = {
        URLSession(configuration: .default, delegate: self, delegateQueue: OperationQueue.main)
    }()

    private var responseOkFilter: HttpSuccessFilter
    private var responseFilters: [Int: HttpResponseFilter] = [:]

    public var delegate: DataManagerDelegate?
    
    public convenience init(host: Host, responseOkFilter: HttpSuccessFilter, responseFilters: [Int: HttpResponseFilter] = [:], timeout: TimeInterval = 10.0) {
        self.init(hostCallback: { host }, responseOkFilter: responseOkFilter, responseFilters: responseFilters, timeout: timeout)
    }
    
    public required init(hostCallback: @escaping HostCallback, responseOkFilter: HttpSuccessFilter, responseFilters: [Int: HttpResponseFilter] = [:], timeout: TimeInterval = 10.0) {
        self.hostCallback = hostCallback
        self.responseOkFilter = responseOkFilter
        self.responseFilters = responseFilters
        self.timeout = timeout
        super.init()
    }

    public func url(username: String = "", password: String = "", path: String) -> String {
        let host = hostCallback()
        var url: String = host.protocol

        if !username.isEmpty || !password.isEmpty {
            url += username + ":" + password + "@"
        }

        url += "\(host.host):\(host.port)\(host.context)"
        url += path

        return url
    }

    @available(iOS 15.0, *)
    @available(macCatalyst 15.0.0, *)
    @discardableResult
    open func request(request r: Request, dataMapper: DataMapper? = nil, sender: Any? = nil) async throws -> Any {
        let request = delegate?.manipulateRequest(request: r) ?? r
        DataManager.logger.info("Requesting \(r)")
        
        guard let req: URLRequest = createUrlRequest(request: request) else {
            throw ResponseError.requestError
        }
        
        let (data, response): (Data, URLResponse)
        do {
            (data, response) = try await urlSession.data(for: req)
        } catch {
            throw self.determineResponseError(url: r.url, error: error)
        }

        guard let response: HTTPURLResponse = response as? HTTPURLResponse else {
            throw ResponseError.typeError
        }
        
        do {
            if response.is2xx {
                return try await self.responseOkFilter.handleResponse(data: data, response: response, dataMapper: dataMapper, sender: sender)
            } else if let filter = self.responseFilters[response.statusCode] {
                return try await filter.handleResponse(request: request, data: data, dataMapper: dataMapper, sender: sender)
            } else {
                throw ResponseError.getErrorForStatusCode(url: r.url, code: response.statusCode)
            }
        } catch {
            throw error
        }
    }

    
    private func createUrlRequest(request: Request) -> URLRequest? {
        if let url: URL = request.constructUrl() {
            var req: URLRequest = URLRequest(url: url)
            req.httpMethod = request.method.rawValue

            if let payload = request.payload {
                req.httpBody = payload
                req.setValue(request.contentType, forHTTPHeaderField: "Content-Type")
                req.setValue("application/json", forHTTPHeaderField: "Accept")
            }

            req.timeoutInterval = request.timeout ?? self.timeout
            
            if let authentication = request.authentication {
                if let headerKey = authentication.headerKey, let headerValue = authentication.headerValue {
                    req.setValue(headerValue, forHTTPHeaderField: headerKey)
                }
            }
            return req
        }
        return nil
    }

    fileprivate func determineResponseError(url: String, error: Error) -> ResponseError {
        let error = error as NSError
        if error.code == NSURLErrorNotConnectedToInternet {
            return ResponseError.networkError(url: url, error: .noInternet)
        } else if error.code == NSURLErrorTimedOut {
            return ResponseError.networkError(url: url, error: .timeout)
        } else if error.code == NSURLErrorCannotConnectToHost {
            return ResponseError.networkError(url: url, error: .serverNotFound)
        } else if error.code == NSURLErrorDataNotAllowed {
            return ResponseError.networkError(url: url, error: .noInternet)
        }
        return ResponseError.networkError(url: url, error: .unknown(error: error))
    }

    public func urlSession(_ session: URLSession, didReceive challenge: URLAuthenticationChallenge, completionHandler: @escaping (URLSession.AuthChallengeDisposition, URLCredential?) -> Void) {
        completionHandler(.useCredential, URLCredential(trust: challenge.protectionSpace.serverTrust!))
    }
}
