//
// Created by Tobias on 14.07.20.
// Copyright (c) 2020 Tobias. All rights reserved.
//

import Foundation

public protocol Authentication
{
    var headerKey: String? { get }

    var headerValue: String? { get }
}

public struct BasicAuthentication: Authentication
{
    private let username: String
    private let password: String

    public init(username: String, password: String) {
        self.username = username
        self.password = password
    }

    public var headerKey: String? {
        "Authorization"
    }
    public var headerValue: String? {
        let credentials = "\(username):\(password)"
        guard let data = credentials.data(using: .utf8) else {
            return nil
        }
        return "Basic \(data.base64EncodedString())"
    }
}

public struct BearerToken: Authentication
{
    private let token: String?
    
    public init(token: String?) {
        self.token = token
    }

    public var headerKey: String? {
        "Authorization"
    }
    public var headerValue: String? {
        guard let token = token else {
            return nil
        }
        return "Bearer \(token)"
    }
}

public class Request: CustomStringConvertible
{
    public enum RequestMethod: String
    {
        case get = "GET"
        case post = "POST"
        case put = "PUT"
        case delete = "DELETE"
    }

    public var url: String
    public var method: RequestMethod
    public var authentication: Authentication?
    public var parameters: [String: String]
    public var payload: Data?

    public var contentType: String?
    
    public var timeout: TimeInterval?

    public init(url: String, method: RequestMethod, authentication: Authentication?, parameters: [String: String] = [:], payload: Data? = nil, contentType: String?, timeout: TimeInterval?) {
        self.url = url
        self.method = method
        self.authentication = authentication
        self.parameters = parameters
        self.payload = payload
        if payload != nil {
            self.contentType = contentType ?? "application/json"
        } else {
            self.contentType = nil
        }
        self.timeout = timeout
    }

    public func constructUrl() -> URL? {
        if var components = URLComponents(string: url) {
            components.queryItems = parameters.map { (key, value) in
                URLQueryItem(name: key, value: value)
            }
            components.percentEncodedQuery = components.percentEncodedQuery?.replacingOccurrences(of: "+", with: "%2B")
            return components.url
        }
        return nil
    }

    public var description: String {
        let filteredParameters: [String: String] = Dictionary(uniqueKeysWithValues: self.parameters.map { (key, value) in
            if key == "password" {
                return (key, "********")
            }
            return (key, value)
        })
        if filteredParameters.isEmpty {
            return "\(method): \(url)"
        } else {
            return "\(method): \(url) \(filteredParameters)"
        }
    }
}
